from __future__ import annotations

import html
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import cli

try:
    from .widget import Widget  # noqa: autoimport


except (ImportError, ModuleNotFoundError):
    pass

if TYPE_CHECKING:  # noqa: autoimport
    from ..asset_types import Section  # noqa: autoimport


@dataclass
class DownloadProgress:
    section: Section
    finished: bool = False
    to_show: bool = False
    amount: int = 0
    progress_value: int = 0
    widget: Any = None

    def request_widget(self):
        from . import userinterface  # noqa: autoimport

        userinterface.UserInterface.request_widget(self)

    def grant_widget(self):
        message = make_message(self.section)
        self.widget = Widget(message)

        self.widget.onCancel = self.onCancel
        self.widget.onOpen = self.onOpen
        self.widget.onClose = self.onClose
        progress = 1 if self.finished else None
        self.set_progress(progress)

        self.widget.exec()

    def onCancel(self):
        self.widget.close()
        self.widget = None
        raise KeyboardInterrupt

    def onOpen(self):
        if self.finished:
            self.show_download()
        else:
            self.to_show = True

    def onClose(self):
        self.widget.close()
        self.widget = None

    def set_progress(self, progress=None):
        if self.widget is None:
            self.request_widget()

        if self.amount and not progress:
            progress = self.progress_value / self.amount
        if progress and self.widget:
            value = int(progress * 100)
            self.widget.progress.setValue(value)
            if progress == 1:
                self.widget.closeButton.setDisabled(False)
                self.widget.enable_keyclose = True

    def add_progress(self, progress):
        self.progress_value += progress
        self.set_progress()

    def add_amount(self, amount):
        self.amount += amount
        self.set_progress()

    def show_download(self):
        self.onClose()
        cli.urlopen(self.section.path)

    def enable_show(self):
        self.set_progress(1)
        if self.to_show:
            self.show_download()
        else:
            self.finished = True


def make_message(section: Section):
    arrow = "\u2B9E "
    message = "\n".join(
        (
            section.course.name,
            *(i * "   " + arrow + part for i, part in enumerate(section.titles)),
            "",
            *(html.unescape(item.display_title) for item in section.items),
        )
    )
    return message



import cli
from plib import Path


class ShortcutMaker:
    def __init__(self):
        self.config_file = Path.HOME / ".xbindkeysrc.scm"
        self.lines = self.config_file.lines
        edge = "; -- scripts --"
        if edge in self.lines:
            self.lines = self.lines[: self.lines.index(edge) + 1]
        self.lines += ["(set-keyboard-bindings", ")"]

    def save_shortcuts(self):
        self.config_file.lines = self.lines
        cli.run("xbindkeys --poll-rc")

    def add_shortcut(self, hotkey, target):
        item = f'\t(list \'({hotkey}) "{target}")' if target else "\n"
        self.lines.insert(-1, item)  # insert before closing parenthese



import io
from dataclasses import dataclass

import dateutil.parser
import pandas as pd

from ..asset_types import Due
from .canvas import assignment, tab


@dataclass
class Checker(assignment.Checker, tab.Checker):
    @classmethod
    def tab_name(cls):
        return "Gradescope"

    @classmethod
    def clean_table(cls, table: pd.DataFrame) -> pd.DataFrame:
        new_column_names = {n: n.split(" (")[0] for n in table.columns}
        table = table.rename(columns=new_column_names)
        table = table[~table["ReleasedDue"].isnull()]
        return table

    def get_content_table(self) -> pd.DataFrame:
        page = self.get_redirected_content()
        with io.StringIO(page) as fp:
            return pd.read_html(fp)[0]

    def extract_row_info(self, row):
        name = self.course.assignment_name(row["Name"])
        date = row["ReleasedDue"]
        return name, date

    def get_items(self):
        table = self.get_content_table()
        table = self.clean_table(table)
        content = [self.extract_row_info(row) for _, row in table.iterrows()]
        return content

    def make_item(self, item):
        name, date = item
        return Due(name, parse_date(date))


def parse_date(date_str):
    end = "Late Due Date: "
    if end in date_str:
        date_str = date_str.split(end)[0].strip()

    start_positions = (
        date_str.rfind(end_keyword, 0, -1) + len(end_keyword)
        for end_keyword in ("AM", "PM")
        if end_keyword in date_str[:-1]
    )
    start_pos = max(start_positions)
    date_str = date_str[start_pos:]
    date = dateutil.parser.parse(date_str)
    return date



import threading


class Threads:
    def __init__(self, target, args=(), kwargs=None):
        if callable(target):
            self.threads = [
                threading.Thread(target=target, args=arg, kwargs=kwargs or {})
                for arg in zip(*args)
            ]
        else:
            self.threads = [
                threading.Thread(target=t, args=args, kwargs=kwargs or {})
                for t in target
            ]

    def start(self):
        for t in self.threads:
            t.start()
        return self

    def join(self):
        for t in self.threads:
            t.join()



from dataclasses import dataclass
from typing import Any, List

from ..utils import Path, time
from . import SaveItem, base


@dataclass
class History(base.Item):
    anon: str
    subject: str
    created: str
    content: str
    uid: str = None
    uid_a: str = None


@dataclass
class Announcement(SaveItem):
    history_size: int
    folders: List[str]
    nr: int
    data: Any
    created: str
    bucket_order: int
    no_answer_followup: int
    change_log: Any
    bucket_name: str
    history: List[History]
    type: str
    tags: List[str]
    tag_good: List[Any]
    unique_views: int
    children: List[Any]
    tag_good_arr: List[str]
    id: str
    config: Any
    status: str
    drafts: Any
    request_instructor: int
    request_instructor_me: bool
    bookmarked: int
    num_favorites: int
    my_favorite: bool
    is_bookmarked: bool
    is_tag_good: bool
    q_edits: Any
    i_edits: Any
    s_edits: Any
    t: int
    default_anonymity: str
    no_answer: Any = None
    save_folder: Path = None

    @property
    def short_title(self):
        return self.history[0].subject

    @property
    def mtime(self):
        return self.history[0].created

    def save(self):
        save_path = self.save_folder / f"piazza_{self.nr}.yaml"
        save_path.yaml = {
            "title": self.short_title,
            "created_at": self.history[0].created,
            "message": self.history[0].content,
        }
        save_path.mtime = time.parse_time(self.mtime)

    def should_download(self):
        return "instructor-note" in self.tags

    @property
    def display_title(self):
        return f"{self.short_title} ({time.export_time(self.mtime)})"

    @property
    def save_id(self):
        return self.id



from .path import Path


def get(course: str):
    root = Path.content_assets / "canvas" / "announ"
    course = course.replace("_", " ")
    for subfolder in root.iterdir():
        if subfolder.name.lower() == course:
            return load_announs(subfolder)


def load_announs(folder: Path):
    announs = [path.yaml for path in folder.iterdir()]
    announs = "<br><hr>".join(
        f"<h2><strong>{announ['title']}</strong><small>&ensp;&ensp;"
        f"{announ['created_at']}"
        f"</small></h2>{announ['message']}" for announ in announs
    )
    return announs



from canvasdl.asset_types.course import Courses
from canvasdl.utils import Path


def get_courses():
    from . import coursemaker  # noqa: autoimport

    courses = Path.courses.yaml or coursemaker.make_courses()
    courses = Courses.from_dict(courses)
    return courses.courses



from dataclasses import dataclass, field

from canvasdl.utils.path import Path

from . import argparser, configmaker


@dataclass
class Config:
    API_URL: str
    API_KEY: str
    one_course_nr: int = None
    update_content: bool = True
    save_content: bool = True
    google_calendar_id: str = None
    drive: dict[str, str] = field(default_factory=dict)
    website: dict[str, str] = field(default_factory=dict)
    piazza: dict[str, str] = field(default_factory=dict)

    @classmethod
    def load(cls):
        args = argparser.get_args()
        if args.configure or not cls.config_file().exists():
            configmaker.make_config()
        config_dict = cls.config_file().yaml
        return Config(**config_dict)

    @classmethod
    def config_file(cls):
        return Path.config if Path.config.exists() else Path.config.encrypted


config = Config.load()



from __future__ import annotations

import typing
from dataclasses import dataclass
from typing import Iterable

from ..utils import Path
from .course import Course

if typing.TYPE_CHECKING:
    from ..asset_types import SaveItem  # noqa: autoimport


@dataclass
class Section:
    path: Path
    course: Course
    items: Iterable[SaveItem]

    @property
    def root(self):
        return Path.school / self.course.name

    @property
    def titles(self):
        for name in self.path.relative_to(self.root).parts:
            yield Path(name).stem

    @property
    def announ(self):
        return self.path.suffix == ".html"



from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from typing import Any

import downloader

from canvasdl import client
from canvasdl.utils.time import parse_time

from ..utils import Path, time
from . import base
from .base import SaveItem


@dataclass
class Asset(base.Item):
    _requester: Any
    id: int


@dataclass
class Item(Asset):
    created_at: str
    created_at_date: datetime
    updated_at: str
    updated_at_date: datetime
    lock_at: None
    unlock_at: None
    locked: bool
    hidden: bool | None
    locked_for_user: bool
    hidden_for_user: bool

    @cached_property
    def mtime(self):
        return parse_time(self.updated_at)


@dataclass
class File(Item):
    uuid: str
    folder_id: int
    display_name: str
    filename: str
    upload_status: str
    content_type: str
    url: str
    size: int
    thumbnail_url: str | None
    modified_at: str
    modified_at_date: datetime
    mime_class: str
    media_entry_id: str | None
    category: str | None = None
    root: Path = None

    @property
    def display_title(self):
        return self.display_name

    @classmethod
    def from_response(cls, response):
        file_info = response.__dict__
        # rename field for parsing
        file_info["content_type"] = file_info.pop("content-type")
        return File.from_dict(file_info)

    @cached_property
    def folder(self):
        folder = client.get_folder(self.folder_id)
        folder = Folder.from_response(folder)
        return folder

    @property
    def folder_path(self):
        return self.root / self.folder.path_name

    def get_full_path(self):
        return self.folder_path / self.display_name

    def save(self):
        path = self.get_full_path()
        downloader.download(self.url, path)
        path.tag = 0
        path.mtime = self.mtime
        path.unpack_if_archive()
        self.folder.set_time(self.root)

    @property
    def save_id(self):
        return self.id


@dataclass
class Folder(Item):
    name: str
    full_name: str
    context_id: int
    context_type: Any
    parent_folder_id: int | None
    position: int | None
    folders_url: str
    files_url: str
    files_count: int
    folders_count: int
    for_submissions: bool
    can_upload: bool

    @property
    def path_name(self):
        folder_name = self.full_name
        for append in ("/", ""):
            folder_name = folder_name.replace(f"course files{append}", "")
        return folder_name

    def set_time(self, root):
        path = root / self.path_name
        path.mtime = self.mtime
        if self.parent:
            self.parent.set_time(root)

    @property
    def parent(self):
        if self.parent_folder_id:
            folder = client.get_folder(self.parent_folder_id)
            folder = Folder.from_response(folder)
        else:
            folder = None
        return folder


@dataclass
class Announcement(Asset, SaveItem):
    title: str
    last_reply_at: str
    last_reply_at_date: datetime
    created_at: str
    created_at_date: datetime
    delayed_post_at: str | None
    posted_at: str
    posted_at_date: datetime
    assignment_id: str | None
    root_topic_id: str | None
    position: int
    podcast_has_student_posts: bool
    discussion_type: str
    lock_at: None
    allow_rating: bool
    only_graders_can_rate: bool
    sort_by_rating: bool
    is_section_specific: bool
    anonymous_state: None
    user_name: str
    discussion_subentry_count: int
    permissions: Any
    require_initial_post: bool | None
    user_can_see_posts: bool
    podcast_url: None
    read_state: str
    unread_count: int
    subscribed: bool
    attachments: Any
    published: bool
    can_unpublish: bool
    locked: bool
    can_lock: bool
    comments_disabled: bool
    author: Any
    html_url: str
    url: str
    pinned: bool
    group_category_id: None
    can_group: bool
    topic_children: Any
    group_topic_children: Any
    context_code: str
    locked_for_user: bool
    message: str
    subscription_hold: Any
    todo_date: None
    delayed_post_at_date: datetime = None

    lock_info: Any = None
    lock_explanation: Any = None
    save_folder: Path = None

    def save(self):
        save_path = self.save_folder / f"{self.position}.yaml"
        save_path.yaml = {
            k: self.__dict__[k] for k in ("title", "created_at", "message")
        }
        save_path.mtime = time.parse_time(self.created_at)

    @property
    def save_id(self):
        return self.id

    @property
    def display_title(self):
        return self.title



import json
from dataclasses import asdict, dataclass

import dacite


@dataclass
class Item:
    @classmethod
    def from_dict(cls, data):
        if isinstance(data, list):
            data = {"items": data}
        return dacite.from_dict(cls, data) 

    @classmethod
    def from_bytes(cls, data):
        return cls.from_dict(json.loads(data)) if data else cls()

    @classmethod
    def from_response(cls, response):
        return cls.from_dict(response.__dict__)

    def dict(self):
        return asdict(self)


@dataclass
class SaveItem(Item):
    @classmethod
    def save(cls):
        raise NotImplementedError

    @property
    def display_title(self):
        raise NotImplementedError

    @property
    def save_id(self):
        raise NotImplementedError

    @classmethod
    def should_download(cls):
        return True



from dataclasses import dataclass
from datetime import datetime

from ..utils import calendar, time
from .base import SaveItem


@dataclass
class Due(SaveItem):
    name: str
    date: datetime

    @property
    def display_title(self):
        return f"{self.name} ({time.export_datetime(self.date)})"

    def save(self):
        calendar.add_todo(self.name, self.date)

    @property
    def save_id(self):
        return self.name



from __future__ import annotations

import typing

import cli
from rich.prompt import Confirm, Prompt

from . import configchecker
from .path import Path

if typing.TYPE_CHECKING:
    from canvasdl.asset_types import Course  # noqa: autoimport


DEFAULT_API_URL = "https://courseworks2.columbia.edu"


def make_config():
    config = get_config_info()
    question = (
        "Do you want to save your config in an encrypted file? "
        "(you need ksshaskpass and gpg installed for this)"
    )
    config_file = (
        Path.config.encrypted if Confirm().ask(question, default=False) else Path.config
    )
    config_file.yaml = config


def get_config_info():
    config = {}

    config = ask_value(
        config,
        "API_URL",
        "Canvas url of your school",
        DEFAULT_API_URL,
        lambda url: configchecker.is_valid_url(url, ""),
    )

    config = ask_value(
        config,
        "API_KEY",
        "Your Canvas API key",
        None,
        lambda key: configchecker.is_valid_key(config["API_URL"], key),
    )

    question = (
        "Do you want to synchronize all important deadlines with your google calendar?"
    )
    sync_calendar = Confirm().ask(question, default=True)

    if sync_calendar:
        dummy_name = "CREDS"
        if not configchecker.google_calendar_credentials_valid():
            ask_value(
                config,
                dummy_name,
                "Please obtain a google calendar credentials file (instructions in"
                f" Readme) and save it to {Path.calendar_credentials} (press enter when"
                " done)",
                None,
                lambda response: configchecker.google_calendar_credentials_valid(),
            )
            config.pop(dummy_name)

        if "google_calendar_id" not in config:
            question = (
                "Do you want to use a custom calendar id? (only relevant if your"
                " account has multiple calendars)"
            )
            if Confirm().ask(question, default=False):
                config = ask_value(
                    config,
                    "google_calendar_id",
                    "Id of your google calendar",
                    None,
                    lambda id: configchecker.google_calendar_credentials_valid(id),
                )
    return config


def ask_value(config, name, message, default=None, check_function=None):
    base_message = message
    message_changed = False

    while name not in config:
        prompt = Prompt()
        response = prompt.ask(message, default=default)
        valid_response = check_function is None
        if not valid_response:
            with cli.status(f"Checking {base_message}"):
                valid_response = check_function(response)

        if valid_response:
            config[name] = response
        elif not message_changed:
            message = f"Error: please provide valid value for {message}"
            message_changed = True

    return config



import itertools

import dbus

from .silentpopup import SilentUIHandle

session_bus = dbus.SessionBus()


def get_dbus_interface(name, path, interface):
    obj = session_bus.get_object(name, path)
    return dbus.Interface(obj, dbus_interface=interface)


class UIHandle:
    def __new__(self, title, app_icon_name=""):
        ui_interface = get_dbus_interface(
            "org.kde.kuiserver", "/JobViewServer", "org.kde.JobViewServer"
        )
        own_interface_path = ui_interface.requestView(title, app_icon_name, 0)

        handle = get_dbus_interface(
            "org.kde.kuiserver", own_interface_path, "org.kde.JobViewV2"
        )
        return handle


class Popup:
    def __init__(
        self,
        *args,
        title="Working",
        message="",
        progress_name=None,
        amount=0,
        description=False,
        capture_errors=False,
        silent=False,
    ):
        self.handle = UIHandle(title) if not silent else SilentUIHandle()
        self.handle.setInfoMessage(message)

        # title can be given by first argument as well
        if args and isinstance(args[0], str):
            title, *args = args
        self.iterator = itertools.chain(*args) if args else None

        self.amount = amount
        if self.amount == 0 and args:
            try:
                self.amount = sum([len(arg) for arg in args])
            except TypeError:
                pass

        self.progress_name = progress_name
        self.message = message
        self.progress = 0
        self.finished = False

        if amount != 0:
            self.handle.setPercent(1)
        if description:
            self.handle.setDescriptionField(0, "", self.message)

    def __enter__(self):
        return self

    def __exit__(self, _, exception, tb):
        self.show_progress()
        self.finished = True

        if not exception:
            message = ""
        elif isinstance(exception, KeyboardInterrupt):
            message = "Cancelled"
        else:
            message = "Error"

        self.handle.setInfoMessage("")
        self.handle.terminate(message)

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value
        self.show_progress()

    def increase_progress(self, value=1):
        self._progress += value

    @property
    def percentage(self):
        return 100 * self.progress / self.amount

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, value):
        self._message = value
        self.show_progress_message()

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, value):
        self._amount = value

    def increase_amount(self, value=1):
        self._amount += value

    def show_progress(self):
        if self.amount != 0 and self.amount is not None:
            # Don't go backwards
            percentage = min(self.percentage, 100)
            self.handle.setPercent(percentage)
            self.show_progress_message()

    def show_progress_message(self):
        message = (
            f"{self.message}\n{self.progress}/{self.amount} {self.progress_name}"
            if self.progress_name
            else self.message
        )
        self.handle.setInfoMessage(message)

    def __iter__(self):
        with self:
            for item in self.iterator:
                yield item
                self.progress += 1

    def iterate(self, iterator):
        self.iterator = iterator
        return self.__iter__()



from datetime import datetime

import pytest

from canvasdl.checkers import gradescope

parsing_pairs = [
    (
        (
            "2 weeks, 6 days leftMar 07 at 4:10PMMar 28 "
            "at 4:10PM Late Due Date: Apr 01 at 11"
        ),
        datetime(2023, 3, 28, 16, 10),
    )
]


@pytest.mark.parametrize("date_str, parsed_date", parsing_pairs)
def test_date_parser(date_str, parsed_date):
    assert gradescope.parse_date(date_str) == parsed_date



import pytest
from hypothesis import Verbosity, given, settings, strategies

from canvasdl.utils import config, configchecker
from canvasdl.utils.path import Path


def test_valid_account():
    assert configchecker.is_valid_account(config.API_URL, config.API_KEY)


@settings(max_examples=1, deadline=5000)
@given(api_url=strategies.text())
def test_invalid_url(api_url):
    api_key = config.API_KEY
    assert not configchecker.is_valid_account(api_url, api_key)
    assert not configchecker.is_valid_url(api_url, api_key)


@given(api_string=strategies.text())
@settings(max_examples=1, deadline=5000)
def test_invalid_formatted_url(api_string):
    api_url = f"https://{api_string}.com"
    api_key = config.API_KEY
    assert not configchecker.is_valid_account(api_url, api_key)
    assert not configchecker.is_valid_url(api_url, api_key)


@settings(max_examples=1, deadline=5000, verbosity=Verbosity.verbose)
@given(api_key=strategies.text())
def test_invalid_key(api_key):
    url = config.API_URL
    assert not configchecker.is_valid_account(url, api_key)
    assert not configchecker.is_valid_key(url, api_key)


def test_google_calendar_credentials():
    assert configchecker.google_calendar_credentials_valid()


def test_custom_calendar_id():
    assert configchecker.google_calendar_credentials_valid(config.google_calendar_id)


@pytest.fixture()
def with_google_calendar_credentials_removed():
    creds = Path.calendar_credentials.json
    token = Path.calendar_token.byte_content

    paths = (Path.calendar_credentials, Path.calendar_token)
    for path in paths:
        path.unlink()

    yield
    Path.calendar_credentials.json = creds
    Path.calendar_token.byte_content = token


def test_invalid_google_calendar_credentials(with_google_calendar_credentials_removed):
    assert not configchecker.google_calendar_credentials_valid()



from .widgetui import WidgetUI


class Widget(WidgetUI):
    def __init__(self, message=""):
        super().__init__(message=message)
        self.onCancel = None
        self.onOpen = None
        self.onClose = None
        self.enable_keyclose = False

        self.closeButton.setDisabled(True)
        self.closeButton.clicked.connect(self.close_clicked)
        self.open.clicked.connect(self.open_clicked)
        self.cancel.clicked.connect(self.cancel_clicked)

    def keyPressEvent(self, event):
        if event.key() == 16777220:  # Enter key
            if self.enable_keyclose:
                self.close()

    def close_clicked(self):
        if self.onClose:
            self.onClose()

    def open_clicked(self):
        if self.onOpen:
            self.onOpen()

    def cancel_clicked(self):
        if self.onCancel:
            self.onCancel()



from abc import ABC
from dataclasses import dataclass

from canvasdl.utils import Path, config
from canvasdl.utils.time import export_time, parse_time

from . import base


@dataclass
class Checker(base.Checker, ABC):
    def __post_init__(self):
        super().__post_init__()
        names = ("announ",)
        self.save_folder = Path.content_path(self.course.name, names).with_suffix("")

    @property
    def path(self):
        return super().path / "Announcements.html"

    def export_downloads(self):
        href = Path.announ_css.as_uri()
        style = f'<link href="{href}" rel="stylesheet" />'
        base_tag = f'<base href="{config.API_URL}">'
        title = f"<br><h1>{self.course.name}</h1><hr>"
        html, last_announ_time = self.load_announs()
        content = base_tag + style + title + html
        self.path.text = content
        self.path.mtime = last_announ_time
        self.path.tag = 9999

    def load_announs(self):
        paths = self.save_folder.iterdir()
        paths = sorted(list(paths), key=lambda path: -int(path.mtime))
        announs = [path.yaml for path in paths]
        last_announ_time = parse_time(announs[0]["created_at"])
        announs = "<br><hr>".join(
            f"<h2><strong>{announ['title']}</strong><small>&ensp;&ensp;"
            f"{export_time(announ['created_at'])}"
            f"</small></h2>{announ['message']}"
            for announ in announs
        )
        return announs, last_announ_time
